#include "sgl.h"
#include "sgl_cd.h"
#include "z80.h"

#define MAX_FILE 128

void LoadFile(char *name, Uint8 *location, Uint32 size);
Uint16 makecol8(Uint8 r, Uint8 g, Uint8 b);
void ClsFast(Uint8 *maplocation);
void Video_Decode(void);
void load_color_proms();
void pdt_test_palette();

unsigned char RAM[0x10000];
Uint16 *color_palette, *map_data, *gfx_data;
Uint8 *tile_data, *tile_data2;

Z80_Regs regs;

int Z80_IRQ=-1;
int Z80_Iperiod = 12000;


int DipSwitchSYNC = 0;

int main(int argc,char **argv)
{
slInitSystem(TV_320x480, NULL, 1);
//Generic memory initialization
color_palette = (Uint16 *) VDP2_COLRAM;		//Set the start of the Color RAM
tile_data = (Uint8 *) VDP2_VRAM_A1;		//Set the Video Ram location
tile_data2 = (Uint8 *) VDP2_VRAM_B1;
map_data = (Uint16 *) VDP2_VRAM_A0;		//Set the Map Data location
gfx_data = (Uint16 *) VDP2_VRAM_A0 + 0x2000;	//Set the Gfx Data location

slPrint("Generic Arcade Emulator v0.02", slLocate(1,1));
slPrint("ported by Denis aka Dagolar", slLocate(1,2));
slPrint("Thanks goes to TakaIsSilly, Charles Doty", slLocate(1,3));
slPrint("Jim Skilskyj and Marcel de Kogel.", slLocate(1,4));
slPrint("Without them it would be impossible", slLocate(1,5));
slPrint("Controls:", slLocate(1,7));
slPrint("Right shift - insert coin", slLocate(1,8));
slPrint("START       - p1 start", slLocate(1,9));
slPrint("C           - fire", slLocate(1,10));
slPrint("B           - shield", slLocate(1,11));
slPrint("D-pad       - right/left", slLocate(1,12));
slPrint("PRESS START to continue", slLocate(1,14));
while(TRUE) {
        if(Per_Connect1) {
            PerDigital* pad;

            pad = Smpc_Peripheral + 0;
            if(!(pad->data & PER_DGT_ST)) break;
        }
        slSynch();
}
ClsSlo();
if (Load_ROMs() == 0)
        {
	ClsFast((void *)map_data);			//Maybe this crashes the Saturn...
	ClsFast((void *)gfx_data);			//Maybe this crashes the Saturn...

	// Standard Saturn tile mode stuff...
	slPageNbg1((void *)tile_data2, 0, PNB_1WORD);
	slPlaneNbg1(PL_SIZE_1x1);
	slMapNbg1((void *)gfx_data , (void *)gfx_data , (void *)gfx_data , (void *)gfx_data);
	slPageNbg0((void *)tile_data2, 0, PNB_1WORD);
	slPlaneNbg0(PL_SIZE_1x1);
	slMapNbg0((void *)map_data , (void *)map_data , (void *)map_data , (void *)map_data);
//slScrTransparent(NBG1ON);
slScrAutoDisp(NBG0ON | NBG1ON);


	Video_Decode();
        if (run_machine() != 0)
                {
                slPrint("unable to start emulation", slLocate(1,2));
                machine_exit();
                }
        else slPrint("unable to initialize system", slLocate(1,3));
        machine_exit();
        }
exit(0);
}

int run_machine()
{
Z80_Reset();
Z80();
return (0);
}

int Load_ROMs()
{
slPrint("Loading Roms...", slLocate(1,1));
LoadFile("IC45.BIN", &RAM[0x0000], 1);
LoadFile("IC46.BIN", &RAM[0x0800], 1);
LoadFile("IC47.BIN", &RAM[0x1000], 1);
LoadFile("IC48.BIN", &RAM[0x1800], 1);
LoadFile("IC49.BIN", &RAM[0x2000], 1);
LoadFile("IC50.BIN", &RAM[0x2800], 1);
LoadFile("IC51.BIN", &RAM[0x3000], 1);
LoadFile("IC52.BIN", &RAM[0x3800], 1);
LoadFile("IC39.BIN", &RAM[0x8000], 1);
LoadFile("IC23.BIN", &RAM[0x8800], 1);
LoadFile("IC40.BIN", &RAM[0x9000], 1);
LoadFile("IC24.BIN", &RAM[0x9800], 1);
load_color_proms();
return(0);
}

int machine_exit()
{
slPrint("This will exit the machine", slLocate(1,15));
exit(0);
}

unsigned Z80_RDMEM(dword A)
{
Uint16 data;

data = Smpc_Peripheral[0].data;

if (A >= 0x7800 && A <=0x7bff)

        {
        if(DipSwitchSYNC == 1)
                {
                VideoDraw();
                DipSwitchSYNC = 0;
		slSynch();
                return 128;
                }
                else
                return 0;
        }
else if (A >= 0x7000 && A <= 0x73FF)
	{
	byte hexdata = 0xff;
	if ((data & PER_DGT_TR) == 0)hexdata &= ~(1<<0);  // Right shift = coin 1
        if ((data & PER_DGT_ST) == 0)hexdata &= ~(1<<1);  // START button = 1p start
	if ((data & PER_DGT_TC) == 0)hexdata &= ~(1<<4); // C button = fire
	if ((data & PER_DGT_KR) == 0)hexdata &= ~(1<<5); // D-Right = right
	if ((data & PER_DGT_KL) == 0)hexdata &= ~(1<<6); // D-Left = left
	if ((data & PER_DGT_TB) == 0)hexdata &= ~(1<<7); // B button = shield
	return hexdata; 
	}
        else if (A < 0x8000)
        return RAM[A];
        slPrint("WARNING! Reading from", slLocate(1,16));
        return RAM[A];
}

void Video_Decode(void)
{
// This function aparently is Saturn-friendly, ie. It turns data into 8x8 tile info.
// We better get some neat stuff out of this.
// Convert the graphics ROMs to a bitmap format

unsigned int line1,line2,bit1,bit2;
unsigned int i,j,c;
Uint8 lookup=0;

for(c=0;c<0x100;c++)     // do a char
	    {
	    for(i=0;i<8;i++) //do a ver. line
	       {
	       line1=RAM[0x8000+(c*8)+7-i];
	       line2=RAM[0x9000+(c*8)+7-i];
		    for(j=0;j<8;j++) //do a hor. line
		       {
		       bit1=((line1&(0x01<<j))>>j);
		       bit2=((line2&(0x01<<j))>>j);
		       lookup=(((c%256)>>5)
			 | (bit1<<3) | (bit2<<4)
			 | (!((c&0xFF00)&&1)<<5));

			//This is suited to Saturn, ie. total conversion of tile data.
			  tile_data[(c*0x40)+(j*8)+i] = (Uint8) lookup;
		       }
	       }
	    }



for(c=0x100;c<0x200;c++)     // do a char
	    {
	    for(i=0;i<8;i++) //do a ver. line
	       {
	       line1=RAM[0x8000+(c*8)+7-i];
	       line2=RAM[0x9000+(c*8)+7-i];
		    for(j=0;j<8;j++) //do a hor. line
		       {
		       bit1=((line1&(0x01<<j))>>j);
		       bit2=((line2&(0x01<<j))>>j);
		       lookup=(((c%256)>>5)
			 | (bit1<<3) | (bit2<<4)
			 | (!((c&0xFF00)&&1)<<5));

			//This is suited to Saturn, ie. total conversion of tile data.
			  tile_data2[((c-0x100)*0x40)+(j*8)+i] = (Uint8) lookup;
		       }
	       }
	    }


}

void load_color_proms()
{
//This needed some modification, random file access is not the best feature of Saturn...

int i=0;
int r=0,g=0,b=0;


unsigned char buffer[0x200];
unsigned char temp[0x800];

slPrint("Loading ColorROM...", slLocate(1,2));

LoadFile("IC41_A.BIN", &temp[0x000], 1);
for (i = 0; i < 0x100; i++)
	buffer[i] = temp[i];
LoadFile("IC40_B.BIN", &temp[0x000], 1);
for (i = 0; i < 0x100; i++)
	buffer[i+0x100] = temp[i];

slPrint("Converting Palletes...", slLocate(1,3));

for(i=0;i<128;i++)
	{

//I'll consider this right, don't really care about the color format.
//It gives good colors, wai ^o^

	    r=((((buffer[i+0x100]>>0)&0x01))|((((buffer[i]>>0)&0x01))<<1))*85;
	    g=((((buffer[i+0x100]>>2)&0x01))|((((buffer[i]>>2)&0x01))<<1))*85;
	    b=((((buffer[i+0x100]>>1)&0x01))|((((buffer[i]>>1)&0x01))<<1))*85;
	    
	    color_palette[i]=makecol8(r,g,b);

	    if((buffer[i+0x100]==0)&&(buffer[i]==0)) color_palette[i]=0;

	}

	for(i=128;i<256;i++) // The rest are white
	{
	  color_palette[(i)] = makecol8(255,255,255);
	}
}

Uint16 makecol8(Uint8 r, Uint8 g, Uint8 b){
//This function turns a 24-bit pallete info to a 15-bit data suitable to Saturn. Should be a macro.
//Maybe a part of a future Allegro to Saturn pseudo library? ^o^

return ((r >> 3)+((g >> 3) << 5)+((b >> 3) << 10));
}

void pdt_test_palette()
	{
// Used for debugging... Saturn format.
    int i=0;
    int j=0;
    int k=0;
     
    for(i=0;i<256;i++)     //do the map...
	{
	map_data[i] = (Uint16) (i << 1);
	}
    }

void Z80_WRMEM(dword A,byte V)
{
if (A < 0x3000)
	{
	return;
	}
	else if(A > 0x3fff && A < 0x8000)
	RAM[A] = V;
return;
}

int InterruptsEnabled(void)
{
return;
}

byte Z80_In(byte Port)
{
return(0);
}
void Z80_Out(byte Port,byte Value)
{
return;
}
void Z80_Reti(void)
{
return;
}
void Z80_Retn(void)
{
return;
}

int Z80_Interrupt(void)
{
DipSwitchSYNC = 1;
return(Z80_IGNORE_INT);
}

void Z80_Patch(Z80_Regs *Regs)
{ 
slPrint("Blank Patch", slLocate(1,16));
return;
}

void Debug()
{
slPrint("Blank Debugger", slLocate(1,16));
return;
}

void ClsFast(Uint8 *maplocation){

//This version clears any tilemap.
int i;
Uint8 *ASCITile;

ASCITile = (Uint8 *) maplocation;
for(i = 0; i <= 0x4000; i++) *(ASCITile++) = 0x00;
}


int ClsSlo(){
int x2,y2;
for(x2 = 0; x2 <= 40; x2++){
  for(y2 = 0; y2 <= 40; y2++){
    slPrint(" ", slLocate(x2,y2));
    }
  }
}

int VideoDraw(dword A)
{
//The big trick here is to make the system think we're using the standard tile format....

int x,y;
Uint8 c;

	for(y = 0; y < 32; y++)
		{
		int pos = 0x4000+32*(26-1)+y;
		int pos2 = 0x4800+32*(26-1)+y;
       		for (x = 0; x < 26; x++)
			{
			map_data[(y*64)+x] = RAM[pos] << 1;
			gfx_data[(y*64)+x] = RAM[pos2] << 1;
			pos -= 32;
                        pos2 -= 32;
	         	}
                }
return(0);
}


void LoadFile(char *name, Uint8 *location, Uint32 size)
{

Sint32 dirwork[SLCD_WORK_SIZE(MAX_FILE)];
Sint32 ndir;
CDHN cdhn;
CDKEY key[2];
CDBUF buf[2];
Sint32 stat;
Sint32 len[2];

ndir = slCdInit(MAX_FILE, dirwork);
key[0].cn = key[0].sm = key[0].ci = CDKEY_NONE;
key[1].cn = CDKEY_TERM;
cdhn = slCdOpen(name, key);
buf[0].type = CDBUF_COPY;
buf[0].trans.copy.addr = location;
buf[0].trans.copy.unit = CDBUF_FORM1;
buf[0].trans.copy.size = size;
buf[1].type = CDBUF_TERM;

slCdLoadFile(cdhn, buf);
while (1) {
slSynch();
stat = slCdGetStatus(cdhn, len);
if (stat == CDSTAT_COMPLETED) break;
}
}
